@@ -26,6 +26,7 @@
 
 use crate::Error;
 use crate::Result;
+use std::cmp;
 
 use crate::frame;
 
@@ -39,7 +40,7 @@ use smallvec::SmallVec;
 /// Used to calculate the cap for the queue of retired connection IDs for which
 /// a RETIRED_CONNECTION_ID frame have not been sent, as a multiple of
 /// `active_conn_id_limit` (see RFC 9000, section 5.1.2).
-const RETIRED_CONN_ID_LIMIT_MULTIPLIER: usize = 3;
+const RETIRED_CONN_ID_LIMIT_MULTIPLIER: u64 = 3;
 
 #[derive(Default)]
 struct BoundedConnectionIdSeqSet {
@@ -71,10 +72,6 @@ impl BoundedConnectionIdSeqSet {
         self.inner.remove(e)
     }
 
-    fn front(&self) -> Option<u64> {
-        self.inner.iter().next().copied()
-    }
-
     fn is_empty(&self) -> bool {
         self.inner.is_empty()
     }
@@ -286,14 +283,16 @@ impl ConnectionIdentifiers {
             },
         );
 
+        // Guard against overflow.
+        let value =
+            (destination_conn_id_limit as u64) * RETIRED_CONN_ID_LIMIT_MULTIPLIER;
+        let size = cmp::min(usize::MAX as u64, value) as usize;
         // Because we already inserted the initial SCID.
         let next_scid_seq = 1;
         ConnectionIdentifiers {
             scids,
             dcids,
-            retire_dcid_seqs: BoundedConnectionIdSeqSet::new(
-                destination_conn_id_limit * RETIRED_CONN_ID_LIMIT_MULTIPLIER,
-            ),
+            retire_dcid_seqs: BoundedConnectionIdSeqSet::new(size),
             next_scid_seq,
             source_conn_id_limit,
             zero_length_scid,
@@ -304,7 +303,7 @@ impl ConnectionIdentifiers {
     /// Sets the maximum number of source connection IDs our peer allows us.
     pub fn set_source_conn_id_limit(&mut self, v: u64) {
         // Bound conn id limit so our scids queue sizing is valid.
-        let v = std::cmp::min(v, (usize::MAX / 2) as u64) as usize;
+        let v = cmp::min(v, (usize::MAX / 2) as u64) as usize;
 
         // It must be at least 2.
         if v >= 2 {
@@ -757,14 +756,15 @@ impl ConnectionIdentifiers {
         self.advertise_new_scid_seqs.front().copied()
     }
 
-    /// Gets a destination Connection IDs's sequence number that need to send
-    /// RETIRE_CONNECTION_ID frames.
+    /// Returns a copy of the set of destination Connection IDs's sequence
+    /// numbers to send RETIRE_CONNECTION_ID frames.
     ///
-    /// If `Some`, it always returns the same value until it has been removed
-    /// using `mark_retire_dcid_seq`.
+    /// Note that the set includes sequence numbers at the time the copy was
+    /// created. To account for newly inserted or removed sequence numbers, a
+    /// new copy needs to be created.
     #[inline]
-    pub fn next_retire_dcid_seq(&self) -> Option<u64> {
-        self.retire_dcid_seqs.front()
+    pub fn retire_dcid_seqs(&self) -> HashSet<u64> {
+        self.retire_dcid_seqs.inner.clone()
     }
 
     /// Returns true if there are new source Connection IDs to advertise.
@@ -828,7 +828,7 @@ impl ConnectionIdentifiers {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::testing::create_cid_and_reset_token;
+    use crate::test_utils::create_cid_and_reset_token;
 
     #[test]
     fn ids_new_scids() {
@@ -927,12 +927,12 @@ mod tests {
         assert_eq!(ids.available_dcids(), 1);
         assert_eq!(ids.dcids.len(), 2);
         assert!(ids.has_retire_dcids());
-        assert_eq!(ids.next_retire_dcid_seq(), Some(0));
+        assert_eq!(ids.retire_dcid_seqs().iter().next(), Some(&0));
 
         // Fake RETIRE_CONNECTION_ID sending.
         let _ = ids.mark_retire_dcid_seq(0, false);
         assert!(!ids.has_retire_dcids());
-        assert_eq!(ids.next_retire_dcid_seq(), None);
+        assert_eq!(ids.retire_dcid_seqs().iter().next(), None);
 
         // Now tries to experience CID retirement. If the server tries to remove
         // non-existing DCIDs, it fails.
@@ -947,13 +947,13 @@ mod tests {
         ids.link_dcid_to_path_id(2, 0).unwrap();
         assert_eq!(ids.available_dcids(), 0);
         assert!(ids.has_retire_dcids());
-        assert_eq!(ids.next_retire_dcid_seq(), Some(1));
+        assert_eq!(ids.retire_dcid_seqs().iter().next(), Some(&1));
         assert_eq!(ids.dcids.len(), 1);
 
         // Fake RETIRE_CONNECTION_ID sending.
         let _ = ids.mark_retire_dcid_seq(1, false);
         assert!(!ids.has_retire_dcids());
-        assert_eq!(ids.next_retire_dcid_seq(), None);
+        assert_eq!(ids.retire_dcid_seqs().iter().next(), None);
 
         // Trying to remove the last DCID triggers an error.
         assert_eq!(ids.retire_dcid(2), Err(Error::OutOfIdentifiers));
