@@ -24,8 +24,6 @@
 // NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 // SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-use debug_panic::debug_panic;
-use std::time::Duration;
 use std::time::Instant;
 
 use self::recovery::Acked;
@@ -38,8 +36,6 @@ use crate::recovery::CongestionControlAlgorithm;
 use crate::StartupExit;
 use crate::StartupExitReason;
 
-pub const PACING_MULTIPLIER: f64 = 1.25;
-
 pub struct SsThresh {
     // Current slow start threshold.  Defaults to usize::MAX which
     // indicates we're still in the initial slow start phase.
@@ -78,7 +74,7 @@ impl SsThresh {
                 // In normal slow start, attribute the exit to loss.
                 StartupExitReason::Loss
             };
-            self.startup_exit = Some(StartupExit::new(ssthresh, reason));
+            self.startup_exit = Some(StartupExit::new(ssthresh, None, reason));
         }
         self.ssthresh = ssthresh;
     }
@@ -93,9 +89,6 @@ pub struct Congestion {
     // HyStart++.
     pub(crate) hystart: hystart::Hystart,
 
-    // Pacing.
-    pub(crate) pacer: pacer::Pacer,
-
     // RFC6937 PRR.
     pub(crate) prr: prr::PRR,
 
@@ -103,12 +96,6 @@ pub struct Congestion {
     // transmitted together.
     send_quantum: usize,
 
-    // BBR state.
-    bbr_state: bbr::State,
-
-    // BBRv2 state.
-    bbr2_state: bbr2::State,
-
     pub(crate) congestion_window: usize,
 
     pub(crate) ssthresh: SsThresh,
@@ -123,8 +110,6 @@ pub struct Congestion {
 
     pub(crate) delivery_rate: delivery_rate::Rate,
 
-    initial_rtt: Duration,
-
     /// Initial congestion window size in terms of packet count.
     pub(crate) initial_congestion_window_packets: usize,
 
@@ -157,8 +142,6 @@ impl Congestion {
 
             lost_count: 0,
 
-            initial_rtt: recovery_config.initial_rtt,
-
             initial_congestion_window_packets: recovery_config
                 .initial_congestion_window_packets,
 
@@ -170,19 +153,7 @@ impl Congestion {
 
             hystart: hystart::Hystart::new(recovery_config.hystart),
 
-            pacer: pacer::Pacer::new(
-                recovery_config.pacing,
-                initial_congestion_window,
-                0,
-                recovery_config.max_send_udp_payload_size,
-                recovery_config.max_pacing_rate,
-            ),
-
             prr: prr::PRR::default(),
-
-            bbr_state: bbr::State::new(),
-
-            bbr2_state: bbr2::State::new(),
         };
 
         (cc.cc_ops.on_init)(&mut cc);
@@ -208,10 +179,6 @@ impl Congestion {
         self.send_quantum
     }
 
-    pub(crate) fn set_pacing_rate(&mut self, rate: u64, now: Instant) {
-        self.pacer.update(self.send_quantum, rate, now);
-    }
-
     pub(crate) fn congestion_window(&self) -> usize {
         self.congestion_window
     }
@@ -223,7 +190,7 @@ impl Congestion {
     #[allow(clippy::too_many_arguments)]
     pub(crate) fn on_packet_sent(
         &mut self, bytes_in_flight: usize, sent_bytes: usize, now: Instant,
-        pkt: &mut Sent, rtt_stats: &RttStats, bytes_lost: u64, in_flight: bool,
+        pkt: &mut Sent, bytes_lost: u64, in_flight: bool,
     ) {
         if in_flight {
             self.update_app_limited(
@@ -242,16 +209,7 @@ impl Congestion {
             }
         }
 
-        // Pacing: Set the pacing rate if CC doesn't do its own.
-        if !(self.cc_ops.has_custom_pacing)() && rtt_stats.has_first_rtt_sample {
-            let rate = PACING_MULTIPLIER * self.congestion_window as f64 /
-                rtt_stats.smoothed_rtt.as_secs_f64();
-            self.set_pacing_rate(rate as u64, now);
-        }
-
-        self.schedule_next_packet(now, sent_bytes);
-
-        pkt.time_sent = self.get_packet_send_time();
+        pkt.time_sent = now;
 
         // bytes_in_flight is already updated. Use previous value.
         self.delivery_rate
@@ -279,27 +237,6 @@ impl Congestion {
             rtt_stats,
         );
     }
-
-    fn schedule_next_packet(&mut self, now: Instant, packet_size: usize) {
-        // Don't pace in any of these cases:
-        //   * Packet contains no data.
-        //   * The congestion window is within initcwnd.
-
-        let in_initcwnd = self.congestion_window <
-            self.max_datagram_size * self.initial_congestion_window_packets;
-
-        let sent_bytes = if !self.pacer.enabled() || in_initcwnd {
-            0
-        } else {
-            packet_size
-        };
-
-        self.pacer.send(sent_bytes, now);
-    }
-
-    pub(crate) fn get_packet_send_time(&self) -> Instant {
-        self.pacer.next_time()
-    }
 }
 
 pub(crate) struct CongestionControlOps {
@@ -332,8 +269,6 @@ pub(crate) struct CongestionControlOps {
 
     pub rollback: fn(r: &mut Congestion) -> bool,
 
-    pub has_custom_pacing: fn() -> bool,
-
     #[cfg(feature = "qlog")]
     pub state_str: fn(r: &Congestion, now: Instant) -> &'static str,
 
@@ -348,12 +283,11 @@ impl From<CongestionControlAlgorithm> for &'static CongestionControlOps {
         match algo {
             CongestionControlAlgorithm::Reno => &reno::RENO,
             CongestionControlAlgorithm::CUBIC => &cubic::CUBIC,
-            CongestionControlAlgorithm::BBR => &bbr::BBR,
-            CongestionControlAlgorithm::BBR2 => &bbr2::BBR2,
-            CongestionControlAlgorithm::Bbr2Gcongestion => {
-                debug_panic!("legacy implementation, not gcongestion");
-                &bbr2::BBR2
-            },
+            // Bbr2Gcongestion is routed to the congestion implementation in
+            // the gcongestion directory by Recovery::new_with_config;
+            // LegacyRecovery never gets a RecoveryConfig with the
+            // Bbr2Gcongestion algorithm.
+            CongestionControlAlgorithm::Bbr2Gcongestion => unreachable!(),
         }
     }
 }
@@ -372,7 +306,7 @@ mod tests {
     #[test]
     fn ssthresh_in_css() {
         let expected_startup_exit =
-            StartupExit::new(1000, StartupExitReason::PersistentQueue);
+            StartupExit::new(1000, None, StartupExitReason::PersistentQueue);
         let mut ssthresh: SsThresh = Default::default();
         ssthresh.update(1000, true);
         assert_eq!(ssthresh.get(), 1000);
@@ -391,7 +325,7 @@ mod tests {
     #[test]
     fn ssthresh_in_slow_start() {
         let expected_startup_exit =
-            StartupExit::new(1000, StartupExitReason::Loss);
+            StartupExit::new(1000, None, StartupExitReason::Loss);
         let mut ssthresh: SsThresh = Default::default();
         ssthresh.update(1000, false);
         assert_eq!(ssthresh.get(), 1000);
@@ -408,12 +342,9 @@ mod tests {
     }
 }
 
-mod bbr;
-mod bbr2;
 mod cubic;
 mod delivery_rate;
 mod hystart;
-pub(crate) mod pacer;
 mod prr;
 pub(crate) mod recovery;
 mod reno;
