@@ -107,8 +107,12 @@ struct Params {
     probe_bw_probe_down_pacing_gain: f32,
     probe_bw_default_pacing_gain: f32,
 
+    /// cwnd_gain for probe bw phases other than ProbeBW_UP
     probe_bw_cwnd_gain: f32,
 
+    /// cwnd_gain for ProbeBW_UP
+    probe_bw_up_cwnd_gain: f32,
+
     // PROBE_UP parameters.
     probe_up_ignore_inflight_hi: bool,
 
@@ -126,6 +130,9 @@ struct Params {
 
     probe_rtt_duration: Duration,
 
+    probe_rtt_pacing_gain: f32,
+    probe_rtt_cwnd_gain: f32,
+
     // Parameters used by multiple modes.
     /// The initial value of the max ack height filter's window length.
     initial_max_ack_height_filter_window: usize,
@@ -160,7 +167,9 @@ struct Params {
     /// STARTUP.
     decrease_startup_pacing_at_end_of_round: bool,
 
-    /// Avoid Overestimation in Bandwidth Sampler with ack aggregation
+    /// Avoid Overestimation in Bandwidth Sampler with ack aggregation.
+    /// This is an old experiment that we have found to under-perform the
+    /// algorithm described in the spec.  Use is not recommended.
     enable_overestimate_avoidance: bool,
 
     /// If true, apply the fix to A0 point selection logic so the
@@ -168,6 +177,7 @@ struct Params {
     /// google/quiche implementation.
     choose_a0_point_fix: bool,
 
+    /// Controls the behavior of BBRAdaptLowerBoundsFromCongestion().
     bw_lo_mode: BwLoMode,
 
     /// Determines whether app limited rounds with no bandwidth growth count
@@ -179,6 +189,24 @@ struct Params {
     /// the initial pacing rate, which is calculated by dividing the
     /// initial cwnd by the first RTT estimate.
     initial_pacing_rate_bytes_per_second: Option<u64>,
+
+    /// If true, scale the pacing rate when updating mss when doing pmtud.
+    scale_pacing_rate_by_mss: bool,
+
+    /// Disable `has_stayed_long_enough_in_probe_down` which can cause ProbeDown
+    /// to exit early.
+    disable_probe_down_early_exit: bool,
+
+    /// Set the expected send time for packets when using BBR to `now`
+    /// instead of `get_next_release_time()`.  Setting the time based
+    /// on `get_next_release_time()` can result in artificially low
+    /// RTT measurements due to the pacer's use of burst_tokens to
+    /// make up for lost time.  BBR has significant problems when
+    /// minRTT is under estimated, so it is better to have the RTT be
+    /// slightly over estimated.  The pacer can only schedule packets
+    /// 1/8th of an RTT into the future, so the error introduced by
+    /// setting `time_sent` to `now` is bounded.
+    time_sent_set_to_now: bool,
 }
 
 impl Params {
@@ -202,6 +230,7 @@ impl Params {
         apply_override!(startup_cwnd_gain);
         apply_override!(startup_pacing_gain);
         apply_override!(full_bw_threshold);
+        apply_override!(startup_full_bw_rounds);
         apply_override!(startup_full_loss_count);
         apply_override!(drain_cwnd_gain);
         apply_override!(drain_pacing_gain);
@@ -211,11 +240,17 @@ impl Params {
         apply_override!(probe_bw_probe_up_pacing_gain);
         apply_override!(probe_bw_probe_down_pacing_gain);
         apply_override!(probe_bw_cwnd_gain);
+        apply_override!(probe_bw_up_cwnd_gain);
+        apply_override!(probe_rtt_pacing_gain);
+        apply_override!(probe_rtt_cwnd_gain);
         apply_override!(max_probe_up_queue_rounds);
         apply_override!(loss_threshold);
         apply_override!(use_bytes_delivered_for_inflight_hi);
         apply_override!(decrease_startup_pacing_at_end_of_round);
         apply_override!(ignore_app_limited_for_no_bandwidth_growth);
+        apply_override!(scale_pacing_rate_by_mss);
+        apply_override!(disable_probe_down_early_exit);
+        apply_override!(time_sent_set_to_now);
         apply_optional_override!(initial_pacing_rate_bytes_per_second);
 
         if let Some(custom_value) = custom_bbr_settings.bw_lo_reduction_strategy {
@@ -259,7 +294,9 @@ const DEFAULT_PARAMS: Params = Params {
 
     probe_bw_default_pacing_gain: 1.0,
 
-    probe_bw_cwnd_gain: 2.25, // BBRv3
+    probe_bw_cwnd_gain: 2.0, // BBRv3
+
+    probe_bw_up_cwnd_gain: 2.25, // BBRv3
 
     probe_up_ignore_inflight_hi: false,
 
@@ -271,6 +308,10 @@ const DEFAULT_PARAMS: Params = Params {
 
     probe_rtt_duration: Duration::from_millis(200),
 
+    probe_rtt_pacing_gain: 1.0,
+
+    probe_rtt_cwnd_gain: 1.0,
+
     initial_max_ack_height_filter_window: 10,
 
     inflight_hi_headroom: 0.15,
@@ -291,22 +332,45 @@ const DEFAULT_PARAMS: Params = Params {
 
     decrease_startup_pacing_at_end_of_round: true,
 
-    enable_overestimate_avoidance: true,
+    enable_overestimate_avoidance: false,
 
     choose_a0_point_fix: false,
 
-    bw_lo_mode: BwLoMode::InflightReduction,
+    bw_lo_mode: BwLoMode::Default,
 
-    ignore_app_limited_for_no_bandwidth_growth: false,
+    ignore_app_limited_for_no_bandwidth_growth: true,
 
     initial_pacing_rate_bytes_per_second: None,
+
+    scale_pacing_rate_by_mss: false,
+
+    disable_probe_down_early_exit: false,
+
+    time_sent_set_to_now: true,
 };
 
 #[derive(Debug, PartialEq)]
 enum BwLoMode {
+    /// Mode that implements the BBRAdaptLowerBoundsFromCongestion()
+    /// behavior described in the BBR RFC draft.
     Default,
+
+    /// BBRAdaptLowerBoundsFromCongestion experiment that reduces
+    /// bw_lo by bytes_lost/min_rtt.
+    ///
+    /// Not recommended.
     MinRttReduction,
+
+    /// BBRAdaptLowerBoundsFromCongestion experiment that reduces
+    /// bw_lo by bw_lo * bytes_lost/inflight.
+    ///
+    /// Not recommended.
     InflightReduction,
+
+    /// BBRAdaptLowerBoundsFromCongestion experiment that reduces
+    /// bw_lo by bw_lo * bytes_lost/cwnd
+    ///
+    /// Not recommended.
     CwndReduction,
 }
 
@@ -439,6 +503,7 @@ impl BBRv2 {
         custom_bbr_params: Option<&BbrParams>,
     ) -> Self {
         let cwnd = initial_congestion_window * max_segment_size;
+
         let params = if let Some(custom_bbr_settings) = custom_bbr_params {
             DEFAULT_PARAMS.with_overrides(custom_bbr_settings)
         } else {
@@ -462,6 +527,10 @@ impl BBRv2 {
         }
     }
 
+    pub fn time_sent_set_to_now(&self) -> bool {
+        self.params.time_sent_set_to_now
+    }
+
     fn on_exit_quiescence(&mut self, now: Instant) {
         if let Some(last_quiescence_start) = self.last_quiescence_start.take() {
             self.mode.do_on_exit_quiescence(
@@ -583,7 +652,6 @@ impl CongestionControl for BBRv2 {
     fn on_packet_sent(
         &mut self, sent_time: Instant, bytes_in_flight: usize,
         packet_number: u64, bytes: usize, is_retransmissible: bool,
-        rtt_stats: &RttStats,
     ) {
         if bytes_in_flight == 0 && self.params.avoid_unnecessary_probe_rtt {
             self.on_exit_quiescence(sent_time);
@@ -596,7 +664,6 @@ impl CongestionControl for BBRv2 {
             packet_number,
             bytes,
             is_retransmissible,
-            rtt_stats,
         );
     }
 
@@ -687,6 +754,10 @@ impl CongestionControl for BBRv2 {
         network_model.bandwidth_estimate()
     }
 
+    fn max_bandwidth(&self) -> Bandwidth {
+        self.mode.network_model().max_bandwidth()
+    }
+
     fn update_mss(&mut self, new_mss: usize) {
         self.cwnd_limits.hi = (self.cwnd_limits.hi as u64 * new_mss as u64 /
             self.mss as u64) as usize;
@@ -696,6 +767,10 @@ impl CongestionControl for BBRv2 {
             (self.cwnd as u64 * new_mss as u64 / self.mss as u64) as usize;
         self.initial_cwnd = (self.initial_cwnd as u64 * new_mss as u64 /
             self.mss as u64) as usize;
+        if self.params.scale_pacing_rate_by_mss {
+            self.pacing_rate =
+                self.pacing_rate * (new_mss as f64 / self.mss as f64);
+        }
         self.mss = new_mss;
     }
 
@@ -712,3 +787,59 @@ impl CongestionControl for BBRv2 {
         self.cwnd_limits.hi = max_cwnd
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use rstest::rstest;
+
+    use super::*;
+
+    #[rstest]
+    fn update_mss(#[values(false, true)] scale_pacing_rate_by_mss: bool) {
+        const INIT_PACKET_SIZE: usize = 1200;
+        const INIT_WINDOW_PACKETS: usize = 10;
+        const MAX_WINDOW_PACKETS: usize = 10000;
+        const INIT_CWND: usize = INIT_WINDOW_PACKETS * INIT_PACKET_SIZE;
+        const MAX_CWND: usize = MAX_WINDOW_PACKETS * INIT_PACKET_SIZE;
+        let initial_rtt = Duration::from_millis(333);
+        let bbr_params = &BbrParams {
+            scale_pacing_rate_by_mss: Some(scale_pacing_rate_by_mss),
+            ..Default::default()
+        };
+
+        const NEW_PACKET_SIZE: usize = 1450;
+        const NEW_CWND: usize = INIT_WINDOW_PACKETS * NEW_PACKET_SIZE;
+        const NEW_MAX_CWND: usize = MAX_WINDOW_PACKETS * NEW_PACKET_SIZE;
+
+        let mut bbr2 = BBRv2::new(
+            INIT_WINDOW_PACKETS,
+            MAX_WINDOW_PACKETS,
+            INIT_PACKET_SIZE,
+            initial_rtt,
+            Some(bbr_params),
+        );
+
+        assert_eq!(bbr2.cwnd_limits.lo, INIT_CWND);
+        assert_eq!(bbr2.cwnd_limits.hi, MAX_CWND);
+        assert_eq!(bbr2.cwnd, INIT_CWND);
+        assert_eq!(
+            bbr2.pacing_rate.to_bytes_per_period(initial_rtt),
+            (2.88499 * INIT_CWND as f64) as u64
+        );
+
+        bbr2.update_mss(NEW_PACKET_SIZE);
+
+        assert_eq!(bbr2.cwnd_limits.lo, NEW_CWND);
+        assert_eq!(bbr2.cwnd_limits.hi, NEW_MAX_CWND);
+        assert_eq!(bbr2.cwnd, NEW_CWND);
+        let pacing_cwnd = if scale_pacing_rate_by_mss {
+            NEW_CWND
+        } else {
+            INIT_CWND
+        };
+        assert_eq!(
+            bbr2.pacing_rate.to_bytes_per_period(initial_rtt),
+            (2.88499 * pacing_cwnd as f64) as u64
+        );
+    }
+}
