@@ -36,8 +36,6 @@ use crate::recovery::gcongestion::bbr::BandwidthSampler;
 use crate::recovery::gcongestion::bbr2::Params;
 use crate::recovery::gcongestion::Bandwidth;
 use crate::recovery::gcongestion::Lost;
-use crate::recovery::rtt::RttStats;
-use crate::recovery::rtt::INITIAL_RTT;
 
 use super::Acked;
 use super::BBRv2CongestionEvent;
@@ -191,7 +189,7 @@ pub(super) struct BBRv2NetworkModel {
 }
 
 impl BBRv2NetworkModel {
-    pub(super) fn new(params: &Params) -> Self {
+    pub(super) fn new(params: &Params, initial_rtt: Duration) -> Self {
         BBRv2NetworkModel {
             min_bytes_in_flight_in_round: usize::MAX,
             inflight_hi_limited_in_round: false,
@@ -206,7 +204,7 @@ impl BBRv2NetworkModel {
                 end_of_round_trip: None,
             },
             min_rtt_filter: MinRttFilter {
-                min_rtt: INITIAL_RTT,
+                min_rtt: initial_rtt,
                 min_rtt_timestamp: Instant::now(),
             },
             max_bandwidth_filter: MaxBandwidthFilter {
@@ -273,7 +271,6 @@ impl BBRv2NetworkModel {
     pub(super) fn on_packet_sent(
         &mut self, sent_time: Instant, bytes_in_flight: usize,
         packet_number: u64, bytes: usize, is_retransmissible: bool,
-        _rtt_stats: &RttStats,
     ) {
         // Updating the min here ensures a more realistic (0) value when flows
         // exit quiescence.
@@ -501,10 +498,11 @@ impl BBRv2NetworkModel {
         // sample_max_bandwidth will be None if the loss is triggered by a timer
         // expiring. Ideally we'd use the most recent bandwidth sample,
         // but bandwidth_latest is safer than None.
-        if congestion_event.sample_max_bandwidth.is_some() {
+        if let Some(sample_max_bandwidth) = congestion_event.sample_max_bandwidth
+        {
             // bandwidth_latest is the max bandwidth for the round, but to allow
             // fast, conservation style response to loss, use the last sample.
-            last_bandwidth = congestion_event.sample_max_bandwidth.unwrap();
+            last_bandwidth = sample_max_bandwidth;
         }
         if self.pacing_gain > params.full_bw_threshold {
             // In STARTUP, `pacing_gain` is applied to `bandwidth_lo` in
@@ -619,6 +617,12 @@ impl BBRv2NetworkModel {
             return true;
         }
 
+        if !congestion_event.last_packet_send_state.is_valid {
+            // last_packet_send_state not available because the
+            // congestion event did not contain any non-ACK frames.
+            return false;
+        }
+
         let ignore_round = self.ignore_app_limited_for_no_bandwidth_growth &&
             congestion_event.last_packet_send_state.is_app_limited;
 
